# -*- coding: utf-8 -*-
"""
    Dummy conftest.py for django_pint.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    https://pytest.org/latest/plugins.html
"""

import django

from pint import UnitRegistry


def pytest_configure(config):
    from django.conf import settings

    custom_ureg = UnitRegistry()
    custom_ureg.define("custom = [custom]")
    custom_ureg.define("kilocustom = 1000 * custom")

    settings.configure(
        DATABASES={
            "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}
        },
        SECRET_KEY="not very secret in tests",
        USE_I18N=True,
        USE_L10N=True,
        # Use common Middleware
        MIDDLEWARE=(
            "django.middleware.common.CommonMiddleware",
            "django.contrib.sessions.middleware.SessionMiddleware",
            "django.contrib.auth.middleware.AuthenticationMiddleware",
            "django.contrib.messages.middleware.MessageMiddleware",
        ),
        INSTALLED_APPS=[
            "django.contrib.auth",
            "django.contrib.admin",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.sites",
            "django.contrib.flatpages",
            "quantityfield",
            "tests.dummyapp",
        ],
        DJANGO_PINT_UNIT_REGISTER=custom_ureg,
    )
    django.setup()



from django.db import models

from quantityfield.fields import (
    BigIntegerQuantityField,
    IntegerQuantityField,
    QuantityField,
)


class HayBale(models.Model):
    name = models.CharField(max_length=20)
    weight = QuantityField("gram")
    weight_int = IntegerQuantityField("gram", blank=True, null=True)
    weight_bigint = BigIntegerQuantityField("gram", blank=True, null=True)


class EmptyHayBale(models.Model):
    name = models.CharField(max_length=20)
    weight = QuantityField("gram", null=True)


class CustomUregHayBale(models.Model):
    # Custom is defined in settings in conftest.py
    custom = QuantityField("custom", null=True)


class ChoicesDefinedInModel(models.Model):
    weight = QuantityField("kilogram", unit_choices=["milligram", "pounds"])



from django import forms
from django.core.exceptions import ValidationError
from django.db import models
from django.utils import formats
from django.utils.translation import gettext_lazy as _

import warnings
from pint import Quantity
from typing import List, Optional

from quantityfield.exceptions import PrecisionLoss
from quantityfield.helper import check_matching_unit_dimension

from .units import ureg
from .widgets import QuantityWidget


def safe_to_int(value):
    """
    Check if a value is an int otherwise warn that it can't be converted
    :param value:
    :return:
    """
    float_value = float(value)
    int_value = int(float_value)
    # We round to compensate for possible float rounding errors
    if round(abs(int_value - float_value), 10) == 0:
        return int_value
    else:
        raise PrecisionLoss(
            _("Possible loss of precision converting value to integer: %s") % value
        )


class QuantityFieldMixin(object):
    to_number_type = NotImplemented

    """A Django Model Field that resolves to a pint Quantity object"""

    def __init__(
        self, base_units: str, *args, unit_choices: Optional[List[str]] = None, **kwargs
    ):
        """
        Create a Quantity field
        :param base_units: Unit description of base unit
        :param unit_choices: If given the possible unit choices with the same
                             dimension like the base_unit
        """
        if not isinstance(base_units, str):
            raise ValueError(
                'QuantityField must be defined with base units, eg: "gram"'
            )

        self.ureg = ureg

        # we do this as a way of raising an exception if some crazy unit was supplied.
        unit = getattr(self.ureg, base_units)  # noqa: F841

        # if we've not hit an exception here, we should be all good
        self.base_units = base_units

        if unit_choices is None:
            self.unit_choices: List[str] = [self.base_units]
        else:
            self.unit_choices = unit_choices

        # Check if all unit_choices are valid
        check_matching_unit_dimension(self.ureg, self.base_units, self.unit_choices)

        super(QuantityFieldMixin, self).__init__(*args, **kwargs)

    @property
    def units(self):
        return self.base_units

    def deconstruct(self):
        name, path, args, kwargs = super(QuantityFieldMixin, self).deconstruct()
        kwargs["base_units"] = self.base_units
        kwargs["unit_choices"] = self.unit_choices
        return name, path, args, kwargs

    def get_prep_value(self, value):
        # we store the value in the base units defined for this field
        if value is None:
            return None

        # Check if not the DeconstructibleUnitRegistry from this module but the default
        # UnitRegistery from pint was used
        if isinstance(value, Quantity) and not isinstance(value, self.ureg.Quantity):
            # Could be fatal if different unit registers are used but we assume
            # the same is used within one project
            # so we might need to implement a check if the UnitRegisters do match
            warnings.warn(
                "Trying to set value from a different unit register for quantityfield. "
                "We assume the naming is equal but best use the same register as for"
                " creating the quantityfield.",
                RuntimeWarning,
            )
            value = value.magnitude * self.ureg(str(value.units))

        if isinstance(value, self.ureg.Quantity):
            to_save = value.to(self.base_units)
            return self.to_number_type(to_save.magnitude)
        return value

    def value_to_string(self, obj) -> str:
        value = self.value_from_object(obj)
        return str(self.get_prep_value(value))

    def from_db_value(self, value, *args, **kwargs):
        if value is None:
            return value
        return self.ureg.Quantity(value * getattr(self.ureg, self.base_units))

    def to_python(self, value):
        if isinstance(value, self.ureg.Quantity):
            return value

        if value is None:
            return None

        return self.ureg.Quantity(value * getattr(self.ureg, self.base_units))

    # TODO: Add tests, understand, add super call if required
    """
    # This code is untested and not documented. It also does not call the super method
    Therefore it is commented out for the moment (even so it is likely required)

    def get_prep_lookup(self, lookup_type, value):

        if lookup_type in ["lt", "gt", "lte", "gte"]:
            if isinstance(value, self.ureg.Quantity):
                v = value.to(self.base_units)
                return v.magnitude
            return value
    """

    def formfield(self, **kwargs):
        defaults = {
            "form_class": self.form_field_class,
            "base_units": self.base_units,
            "unit_choices": self.unit_choices,
        }
        defaults.update(kwargs)
        return super(QuantityFieldMixin, self).formfield(**defaults)


class QuantityFormFieldMixin(object):
    """This formfield allows a user to choose which units they
    wish to use to enter a value, but the value is yielded in
    the base_units
    """

    to_number_type = NotImplemented

    def __init__(self, *args, **kwargs):
        self.ureg = ureg
        self.base_units = kwargs.pop("base_units", None)
        if self.base_units is None:
            raise ValueError(
                "QuantityFormField requires a base_units kwarg of a "
                "single unit type (eg: grams)"
            )
        self.units = kwargs.pop("unit_choices", [self.base_units])
        if self.base_units not in self.units:
            self.units.append(self.base_units)

        check_matching_unit_dimension(self.ureg, self.base_units, self.units)

        kwargs["widget"] = kwargs.get(
            "widget",
            QuantityWidget(base_units=self.base_units, allowed_types=self.units),
        )
        super(QuantityFormFieldMixin, self).__init__(*args, **kwargs)

    def prepare_value(self, value):

        if isinstance(value, Quantity):
            return value.to(self.base_units)
        else:
            return value

    def clean(self, value):
        """
        General idea, first try to extract the correct number like done in the other
        classes and then follow the same procedure as in the django default field
        """
        if isinstance(value, list) or isinstance(value, tuple):
            val = value[0]
            units = value[1]
        else:
            # If no multi widget is used
            val = value
            units = self.base_units

        if val in self.empty_values:
            # Make sure the correct functions are called also in case of empty values
            self.validate(None)
            self.run_validators(None)
            return None

        if units not in self.units:
            raise ValidationError("%(units)s is not a valid choice" % locals())

        if self.localize:
            val = formats.sanitize_separators(value)

        try:
            val = self.to_number_type(val)
        except PrecisionLoss as e:
            raise ValidationError(str(e), code="precision_loss")
        except (ValueError, TypeError):
            raise ValidationError(self.error_messages["invalid"], code="invalid")

        val = self.ureg.Quantity(val * getattr(self.ureg, units)).to(self.base_units)
        self.validate(val.magnitude)
        self.run_validators(val)
        return val


class QuantityFormField(QuantityFormFieldMixin, forms.FloatField):
    to_number_type = float


class QuantityField(QuantityFieldMixin, models.FloatField):
    form_field_class = QuantityFormField
    to_number_type = float


class IntegerQuantityFormField(QuantityFormFieldMixin, forms.IntegerField):
    to_number_type = staticmethod(safe_to_int)


class IntegerQuantityField(QuantityFieldMixin, models.IntegerField):
    form_field_class = IntegerQuantityFormField
    to_number_type = staticmethod(safe_to_int)


class BigIntegerQuantityField(QuantityFieldMixin, models.BigIntegerField):
    form_field_class = IntegerQuantityFormField
    to_number_type = staticmethod(safe_to_int)



import pytest

from django.core.serializers import serialize
from django.db import transaction
from django.test import TestCase

import json
import warnings
from pint import DimensionalityError, UndefinedUnitError, UnitRegistry

from quantityfield.fields import QuantityField
from quantityfield.units import ureg
from tests.dummyapp.models import CustomUregHayBale, EmptyHayBale, HayBale

Quantity = ureg.Quantity


class TestFieldCreate(TestCase):
    def test_sets_units(self):
        test_grams = QuantityField("gram")
        self.assertEqual(test_grams.units, ureg.gram)

    def test_fails_with_unknown_units(self):
        with self.assertRaises(UndefinedUnitError):
            test_crazy_units = QuantityField("zinghie")  # noqa: F841

    def test_base_units_is_required(self):
        with self.assertRaises(TypeError):
            no_units = QuantityField()  # noqa: F841

    def test_base_units_set_with_name(self):
        okay_units = QuantityField(base_units="meter")  # noqa: F841

    def test_base_units_are_invalid(self):
        with self.assertRaises(ValueError):
            wrong_units = QuantityField(None)  # noqa: F841

    def test_unit_choices_must_be_valid_units(self):
        with self.assertRaises(UndefinedUnitError):
            QuantityField(base_units="mile", unit_choices=["gunzu"])

    def test_unit_choices_must_match_base_dimensionality(self):
        with self.assertRaises(DimensionalityError):
            QuantityField(base_units="gram", unit_choices=["meter", "ounces"])


@pytest.mark.django_db
class TestFieldSave(TestCase):
    def setUp(self):
        HayBale.objects.create(
            weight=100, weight_int=100, weight_bigint=100, name="grams"
        )
        HayBale.objects.create(weight=Quantity(10 * ureg.ounce), name="ounce")
        self.lightest = HayBale.objects.create(weight=1, name="lightest")
        self.heaviest = HayBale.objects.create(weight=1000, name="heaviest")
        EmptyHayBale.objects.create(name="Empty")
        CustomUregHayBale.objects.create(custom=5)
        CustomUregHayBale.objects.create(custom=5 * ureg.kilocustom)

    def test_stores_value_in_base_units(self):
        item = HayBale.objects.get(name="ounce")
        self.assertEqual(item.weight.units, "gram")
        self.assertAlmostEqual(item.weight.magnitude, 283.49523125)

    def test_store_integer_loss_of_precision(self):
        with transaction.atomic():
            with self.assertRaisesRegex(ValueError, "loss of precision"):
                HayBale(name="x", weight=0, weight_int=Quantity(10 * ureg.ounce)).save()
        with transaction.atomic():
            with self.assertRaisesRegex(ValueError, "loss of precision"):
                HayBale(
                    name="x", weight=0, weight_bigint=Quantity(10 * ureg.ounce)
                ).save()

    def test_fails_with_incompatible_units(self):
        # we have to wrap this in a transaction
        # fixing a unit test problem
        # http://stackoverflow.com/questions/21458387/transactionmanagementerror-you-cant-execute-queries-until-the-end-of-the-atom
        metres = Quantity(100 * ureg.meter)
        with transaction.atomic():
            with self.assertRaises(DimensionalityError):
                HayBale.objects.create(weight=metres, name="Should Fail")

    def test_accepts_auto_null(self):
        empty = EmptyHayBale.objects.first()
        self.assertIsNone(empty.weight, None)

    def test_accepts_assigned_null(self):
        new = EmptyHayBale()
        new.weight = None
        new.name = "Test"
        new.save()
        self.assertIsNone(new.weight)

    def test_accepts_assigned_float(self):
        new = EmptyHayBale(name="FloatTest")
        new.weight = 707
        new.save()
        obj: EmptyHayBale = EmptyHayBale.objects.last()
        self.assertEqual(obj.name, "FloatTest")
        self.assertEqual(obj.weight.units, "gram")
        self.assertEqual(obj.weight.magnitude, 707)

    def test_accepts_default_pint_unit(self):
        new = EmptyHayBale(name="DefaultPintUnitTest")
        units = UnitRegistry()
        new.weight = 5 * units.kilogram
        # Different Registers so we expect a warning!
        with self.assertWarns(RuntimeWarning):
            new.save()
        obj: EmptyHayBale = EmptyHayBale.objects.last()
        self.assertEqual(obj.name, "DefaultPintUnitTest")
        self.assertEqual(obj.weight.units, "gram")
        self.assertEqual(obj.weight.magnitude, 5000)

    def test_accepts_default_app_unit(self):
        new = EmptyHayBale(name="DefaultAppUnitTest")
        new.weight = 5 * ureg.kilogram
        # Make sure that the correct argument does not raise a warning
        with warnings.catch_warnings(record=True) as w:
            new.save()
        assert len(w) == 0
        obj: EmptyHayBale = EmptyHayBale.objects.last()
        self.assertEqual(obj.name, "DefaultAppUnitTest")
        self.assertEqual(obj.weight.units, "gram")
        self.assertEqual(obj.weight.magnitude, 5000)

    def test_value_stored_as_quantity(self):
        obj = HayBale.objects.first()
        self.assertIsInstance(obj.weight, Quantity)
        self.assertEqual(str(obj.weight), "100.0 gram")
        self.assertEqual(str(obj.weight_int), "100 gram")
        self.assertEqual(str(obj.weight_bigint), "100 gram")

    def test_value_conversion(self):
        obj = HayBale.objects.first()
        ounces = obj.weight.to(ureg.ounce)
        self.assertAlmostEqual(ounces.magnitude, 3.52739619496)
        self.assertEqual(ounces.units, ureg.ounce)

    def test_order_by(self):
        qs = HayBale.objects.all().order_by("weight")
        self.assertEqual(qs[0].name, "lightest")

    def test_comparison_with_number(self):
        qs = HayBale.objects.filter(weight__gt=2)
        self.assertNotIn(self.lightest, qs)

    def test_comparison_with_quantity(self):
        weight = Quantity(20 * ureg.gram)
        qs = HayBale.objects.filter(weight__gt=weight)
        self.assertNotIn(self.lightest, qs)

    def test_comparison_with_quantity_respects_units(self):
        # 1 ounce = 28.34 grams
        weight = Quantity(0.8 * ureg.ounce)
        qs = HayBale.objects.filter(weight__gt=weight)
        self.assertNotIn(self.lightest, qs)

    def test_comparison_is_actually_numeric(self):
        qs = HayBale.objects.filter(weight__gt=1.0)
        self.assertNotIn(self.lightest, qs)

    def tearDown(self):
        HayBale.objects.all().delete()
        EmptyHayBale.objects.all().delete()

    def test_custom_ureg(self):
        obj = CustomUregHayBale.objects.first()
        self.assertIsInstance(obj.custom, ureg.Quantity)
        self.assertEqual(str(obj.custom), "5.0 custom")

        obj = CustomUregHayBale.objects.last()
        self.assertEqual(str(obj.custom), "5000.0 custom")

    def test_serialisation(self):
        serialized = serialize(
            "json",
            [
                HayBale.objects.first(),
            ],
        )
        deserialized = json.loads(serialized)
        obj = deserialized[0]["fields"]
        self.assertEqual(obj["weight"], "100.0")
        self.assertEqual(obj["weight_int"], "100")



# flake8: noqa: F841
from django import forms
from django.test import TestCase

from pint import DimensionalityError, UndefinedUnitError

from quantityfield.fields import IntegerQuantityFormField, QuantityFormField
from quantityfield.units import ureg
from quantityfield.widgets import QuantityWidget
from tests.dummyapp.models import ChoicesDefinedInModel, HayBale

Quantity = ureg.Quantity


class HayBaleForm(forms.ModelForm):
    weight = QuantityFormField(base_units="gram", unit_choices=["ounce", "gram"])
    weight_int = IntegerQuantityFormField(
        base_units="gram", unit_choices=["ounce", "gram"]
    )

    class Meta:
        model = HayBale
        exclude = ["weight_bigint"]


class HayBaleFormDefaultWidgets(forms.ModelForm):
    weight = QuantityFormField(
        base_units="gram", unit_choices=["ounce", "gram"], widget=forms.NumberInput
    )
    weight_int = IntegerQuantityFormField(
        base_units="gram", unit_choices=["ounce", "gram"], widget=forms.NumberInput
    )

    class Meta:
        model = HayBale
        exclude = ["weight_bigint"]


class UnitChoicesDefinedInModelFieldModelForm(forms.ModelForm):
    class Meta:
        model = ChoicesDefinedInModel
        fields = ["weight"]


class NullableWeightForm(forms.Form):
    weight = QuantityFormField(base_units="gram", required=False)


class UnitChoicesForm(forms.Form):
    distance = QuantityFormField(
        base_units="kilometer", unit_choices=["mile", "kilometer", "yard", "feet"]
    )


class TestWidgets(TestCase):
    def test_creates_correct_widget_for_modelform(self):
        form = HayBaleForm()
        self.assertIsInstance(form.fields["weight"], QuantityFormField)
        self.assertIsInstance(form.fields["weight"].widget, QuantityWidget)

    def test_displays_initial_data_correctly(self):
        form = HayBaleForm(
            initial={"weight": Quantity(100 * ureg.gram), "name": "test"}
        )

    def test_clean_yields_quantity(self):
        form = HayBaleForm(
            data={
                "weight_0": 100.0,
                "weight_1": "gram",
                "weight_int_0": 100,
                "weight_int_1": "gram",
                "name": "test",
            }
        )
        self.assertTrue(form.is_valid())
        self.assertIsInstance(form.cleaned_data["weight"], Quantity)

    def test_clean_yields_quantity_in_correct_units(self):
        form = HayBaleForm(
            data={
                "weight_0": 1.0,
                "weight_1": "ounce",
                "weight_int_0": 1,
                "weight_int_1": "ounce",
                "name": "test",
            }
        )
        self.assertTrue(form.is_valid())
        self.assertEqual(str(form.cleaned_data["weight"].units), "gram")
        self.assertAlmostEqual(form.cleaned_data["weight"].magnitude, 28.349523125)

    def test_base_units_is_required_for_form_field(self):
        with self.assertRaises(ValueError):
            field = QuantityFormField()  # noqa: F841

    def test_quantityfield_can_be_null(self):
        form = NullableWeightForm(data={"weight_0": None, "weight_1": None})
        self.assertTrue(form.is_valid())

    def test_validate_units(self):
        form = UnitChoicesForm(data={"distance_0": 100, "distance_1": "ounce"})
        self.assertFalse(form.is_valid())

    def test_base_units_is_included_by_default(self):
        field = QuantityFormField(base_units="mile", unit_choices=["meters", "feet"])
        self.assertIn("mile", field.units)

    def test_widget_field_displays_unit_choices(self):
        form = UnitChoicesForm()
        self.assertListEqual(
            [
                ("mile", "mile"),
                ("kilometer", "kilometer"),
                ("yard", "yard"),
                ("feet", "feet"),
            ],
            form.fields["distance"].widget.widgets[1].choices,
        )

    def test_widget_field_displays_unit_choices_for_model_field_propagation(self):
        form = UnitChoicesDefinedInModelFieldModelForm()
        self.assertListEqual(
            [
                ("milligram", "milligram"),
                ("pounds", "pounds"),
                ("kilogram", "kilogram"),
            ],
            form.fields["weight"].widget.widgets[1].choices,
        )

    def test_unit_choices_must_be_valid_units(self):
        with self.assertRaises(UndefinedUnitError):
            field = QuantityFormField(
                base_units="mile", unit_choices=["gunzu"]
            )  # noqa: F841

    def test_unit_choices_must_match_base_dimensionality(self):
        with self.assertRaises(DimensionalityError):
            field = QuantityFormField(
                base_units="gram", unit_choices=["meter", "ounces"]
            )  # noqa: F841

    def test_widget_display(self):
        bale = HayBale.objects.create(name="Fritz", weight=20)
        form = HayBaleForm(instance=bale)
        html = str(form)
        self.assertIn(
            '<input type="number" name="weight_int_0" step="any" '
            'required id="id_weight_int_0">',
            html,
        )

        self.assertIn('<option value="ounce">ounce</option>', html)

    def test_widget_valid_inputs_with_units(self):
        form = HayBaleForm(
            data={
                "name": "testing",
                "weight_0": "10.3",
                "weight_1": "gram",
                "weight_int_0": "10",
                "weight_int_1": "gram",
            }
        )
        self.assertTrue(form.is_valid())
        self.assertAlmostEqual(form.cleaned_data["weight"].magnitude, 10.3)
        self.assertEqual(form.cleaned_data["weight_int"].magnitude, 10)
        self.assertEqual(str(form.cleaned_data["weight"].units), "gram")
        self.assertEqual(str(form.cleaned_data["weight_int"].units), "gram")

    def test_widget_invalid_float(self):
        form = HayBaleForm(
            data={
                "name": "testing",
                "weight_0": "a",
                "weight_1": "gram",
                "weight_int_0": "10",
                "weight_int_1": "gram",
            }
        )
        self.assertFalse(form.is_valid())
        self.assertIn("weight", form.errors)

    def test_widget_missing_required_input(self):
        form = HayBaleForm(
            data={
                "name": "testing",
                "weight_int_0": "10",
                "weight_int_1": "gram",
            }
        )
        self.assertFalse(form.is_valid())
        self.assertIn("weight", form.errors)

    def test_widget_empty_value_for_required_input(self):
        form = HayBaleForm(
            data={
                "name": "testing",
                "weight_0": "",
                "weight_1": "gram",
                "weight_int_0": "10",
                "weight_int_1": "gram",
            }
        )
        self.assertFalse(form.is_valid())
        self.assertIn("weight", form.errors)

    def test_widget_none_value_set_for_required_input(self):
        form = HayBaleForm(
            data={
                "name": "testing",
                "weight_0": None,
                "weight_1": "gram",
                "weight_int_0": "10",
                "weight_int_1": "gram",
            }
        )
        self.assertFalse(form.is_valid())
        self.assertIn("weight", form.errors)

    def test_widget_single_inputs_with_units(self):
        form = HayBaleFormDefaultWidgets(
            data={
                "name": "testing",
                "weight": "10.3",
                "weight_int": "10",
            }
        )
        self.assertTrue(form.is_valid())
        self.assertAlmostEqual(form.cleaned_data["weight"].magnitude, 10.3)
        self.assertEqual(form.cleaned_data["weight_int"].magnitude, 10)
        self.assertEqual(str(form.cleaned_data["weight"].units), "gram")
        self.assertEqual(str(form.cleaned_data["weight_int"].units), "gram")

    def test_widget_int_precision_loss(self):
        form = HayBaleFormDefaultWidgets(
            data={
                "name": "testing",
                "weight": "10",
                "weight_int": "10.3",
            }
        )
        self.assertFalse(form.is_valid())
        self.assertTrue(form.has_error("weight_int", "precision_loss"))
